--[[ ArenaLive Core Functions: Role Icon Handler
Created by: Vadrak
Creation Date: 31.01.2015
Last Update: "
This basically is a copy-paste of the regular role icon in order to show a custom role texture.
]]--

-- ArenaLive addon Name and localisation table:
local addonName, L = ...;

--[[
**************************************************
******* GENERAL HANDLER SET UP STARTS HERE *******
**************************************************
]]--
local RoleIcon = ArenaLive:ConstructHandler("CustomRoleIcon", true, false, false);
RoleIcon:RegisterEvent("PLAYER_ROLES_ASSIGNED");
RoleIcon:RegisterEvent("GROUP_ROSTER_UPDATE");

local ROLE_TEX_COORDS = {
	["HEALER"] = { 0, 0.5, 0, 0.5},
	["DAMAGER"] = {0.5, 1, 0, 0.5},
	["TANK"] = {0, 0.5, 0.5, 1},
};


--[[
****************************************
****** HANDLER METHODS START HERE ******
****************************************
]]--
function RoleIcon:Update(unitFrame)
	
	local unit = unitFrame.unit;
	local icon = unitFrame[self.name];
	if ( not unit or not icon ) then
		return;
	end
	
	local role = UnitGroupRolesAssigned(unit);
	if ( role and role ~= "NONE" ) then
		icon:SetTexture("Interface\\AddOns\\BGLive\\Textures\\RoleIcons");
		icon:SetTexCoord(unpack(ROLE_TEX_COORDS[role]));
	else
		self:Reset(unitFrame);
	end
end

function RoleIcon:Reset(unitFrame)
	local icon = unitFrame[self.name];
	icon:Hide();
end

function RoleIcon:GetTexture (unitFrame)
	
	local unit = unitFrame.unit;
	if ( not unit ) then
		return nil, 0, 1, 0, 1;
	end

	local role = UnitGroupRolesAssigned(unit);
	if ( role == "TANK" or role == "HEALER" or role == "DAMAGER") then
		return 
	else
		return nil, 0, 1, 0, 1;
	end

end

function RoleIcon:GetShown (unitFrame)
	
	local unit = unitFrame.unit;
	if ( not unit ) then
		return false;
	end

	local role = UnitGroupRolesAssigned(unit);
	if ( role == "TANK" or role == "HEALER" or role == "DAMAGER") then
		return true;
	else
		return false;
	end

end

function RoleIcon:OnEvent(event, ...)
	if ( event == "PLAYER_ROLES_ASSIGNED" ) then
		local unit = "player";
		if ( ArenaLive:IsUnitInUnitFrameCache(unit) ) then
			for id in ArenaLive:GetAffectedUnitFramesByUnit(unit) do
				local unitFrame = ArenaLive:GetUnitFrameByID(id);
				if ( unitFrame[self.name] ) then
					RoleIcon:Update(unitFrame);
				end
			end
		end
	elseif ( event == "GROUP_ROSTER_UPDATE" ) then
		for id, unitFrame in ArenaLive:GetAllUnitFrames() do
			if ( unitFrame[self.name] ) then
				RoleIcon:Update(unitFrame);
			end
		end	
	end
end