local addonName, L = ...;

BGLive.defaults = {
	["BaseIndicator"] = {
		["Direction"] = "RIGHT",
		["IconXOffset"] = 25,
		["IconYOffset"] = 0
	},
	["Cooldown"] =	{
		["ShowText"] = true,
		["StaticSize"] = false,
		["TextSize"] = 8,
	},
	["CCIndicator"] =	{
		["Priorities"] = {
			["defCD"] = 6,
			["offCD"] = 5,
			["stun"] = 4,
			["silence"] = 3,
			["crowdControl"] = 2,
			["root"] = 1,
			["usefulBuffs"] = 0,
			["usefulDebuffs"] = 0,
		},
	},
	["FirstLogin"] = true,
	["FollowTarget"] = true,
	["FrameMover"] = {
		["FrameLock"] = true,
	},
	["ShowScoreBoard"] = true,
	["SideFrames"] = {
		["UnitFrame"] = {
			["Enabled"] = true,
			["TooltipMode"] = "Never",
			["Scale"] = 1,
		},
		["CCIndicator"] = {
			["Enabled"] = true,
		},
		["HealthBar"] = {
			["ColourMode"] = "class",
			["ShowHealPrediction"] = true,
			["ShowAbsorb"] = true,
		},
		["HealthBarText"] = {
			["BarText"] = "%CURR_SHORT% (%PERCENT_SHORT%%)",
			["FontObject"] = "BGLiveSideFrameStatusBarFont",
			["ShowDeadOrGhost"] = true,
			["ShowDisconnect"] = true,
		},
		["PowerBar"] = {
		},
		["PowerBarText"] = {
			["BarText"] = "%CURR_SHORT% (%PERCENT_SHORT%%)",
			["FontObject"] = "BGLiveSideFrameStatusBarFont",
		},
		["Portrait"] = {
			["Type"] = "class",
		},
		["Icon"] = {
			[1] = {
				["Type"] = "trinket",
			},
			[2] = {
				["Type"] = "defensiveCD",
			},
		},
		["NameText"] = {
			["ColourMode"] = "none",
			["FontObject"] = "BGLiveSideFrameNameFont",
		},
	},
	["TeamA"] = {
		["Colour"] = { 0, 0.48, 0.87, 1 },
	},
	["TeamB"] = {
		["Colour"] = { 0.87, 0, 0, 1 },
	},
	["Version"] = "0.0.1a",
};

local baseToFaction = {};

function BGLive:Enable(zoneName, battlefieldStatusID)
	if ( self.enabled and battlefieldStatusID == self.currentID ) then
		return;
	end
	
	-- Register needed events at this point.
	BGLive:RegisterEvent("ARENA_OPPONENT_UPDATE");
	BGLive:RegisterEvent("AL_SPEC_PLAYER_UPDATE");
	BGLive:RegisterEvent("CHAT_MSG_BG_SYSTEM_ALLIANCE");
	BGLive:RegisterEvent("CHAT_MSG_BG_SYSTEM_HORDE");
	BGLive:RegisterEvent("CHAT_MSG_BG_SYSTEM_NEUTRAL");
	BGLive:RegisterEvent("CHAT_MSG_RAID_BOSS_EMOTE");
	BGLive:RegisterEvent("PLAYER_TARGET_CHANGED");
	BGLive:RegisterEvent("BATTLEGROUND_POINTS_UPDATE");
	BGLive:RegisterEvent("WORLD_STATE_UI_TIMER_UPDATE");
	if ( zoneName == L["Alterac Valley"] or zoneName == L["Isle of Conquest"] or zoneName == L["Temple of Kotmogu"] or zoneName == L["Silvershard Mines"] ) then
		BGLive:RegisterEvent("UPDATE_WORLD_STATES");
	end

	self.currentMap = zoneName;
	self.currentID = battlefieldStatusID;
	
	self:EnableCarrierFrames();
	self:EnableSideFrames();
		
	self:UpdateCarrierFrames();
	self:UpdateSideFrames();
		
	BGLiveScoreBoard:UpdatePoints();
	BGLiveScoreBoard:UpdateTimer();
	ArenaLiveSpectatorHideUIButton:Show();
		
	self:Show();
	UIParent:Hide();
	self.enabled = true;
end

function BGLive:Disable()
	BGLive:DisableSideFrames();
	BGLive:DisableCarrierFrames();
	BGLiveScoreBoard:Reset();
	
	if ( not ArenaLiveSpectator:IsEnabled() ) then
		ArenaLiveSpectatorHideUIButton:Hide();
	end

	-- Unregister events:
	BGLive:UnregisterEvent("AL_SPEC_PLAYER_UPDATE");
	BGLive:UnregisterEvent("ARENA_OPPONENT_UPDATE");
	BGLive:UnregisterEvent("CHAT_MSG_BG_SYSTEM_ALLIANCE");
	BGLive:UnregisterEvent("CHAT_MSG_BG_SYSTEM_HORDE");
	BGLive:UnregisterEvent("CHAT_MSG_BG_SYSTEM_NEUTRAL");
	BGLive:UnregisterEvent("CHAT_MSG_RAID_BOSS_EMOTE");
	BGLive:UnregisterEvent("PLAYER_TARGET_CHANGED");
	BGLive:UnregisterEvent("BATTLEGROUND_POINTS_UPDATE");
	BGLive:UnregisterEvent("WORLD_STATE_UI_TIMER_UPDATE");	
	BGLive:UnregisterEvent("UPDATE_WORLD_STATES");
		
	self.worldStateTimerIndex = nil;
	self.currentMap = nil;
	self.enabled = false;
	UIParent:Show();
	self:Hide();
end

function BGLive:GetCurrentMapName()
	return self.currentMap;
end

function BGLive:OnEvent (event, ...)
	local filter = ...;
	if ( event == "ADDON_LOADED" and filter == addonName ) then
		self:InitialiseSideFrames();
		self:InitialiseCarrierFrames();
		BGLiveScoreBoard:Initialise();
		BGLiveChatMessageFrame:Initialise();
		ArenaLive:ConstructHandlerObject(BGLiveBaseIndicator, "BaseIndicator", addonName);
		BGLive:Disable();
	elseif ( event == "ARENA_OPPONENT_UPDATE" ) then
		self:UpdateCarrierFrames();
	elseif ( event == "CHAT_MSG_BG_SYSTEM_ALLIANCE" or event == "CHAT_MSG_BG_SYSTEM_HORDE" or event == "CHAT_MSG_BG_SYSTEM_NEUTRAL" or event == "CHAT_MSG_RAID_BOSS_EMOTE" ) then
		BGLiveChatMessageFrame:OnEvent(event, ...);
	elseif ( event == "PLAYER_TARGET_CHANGED" ) then
		if ( self.enabled ) then
			local database = ArenaLive:GetDBComponent(ArenaLiveSpectator.name);
			if ( database.FollowTarget ) then
				CommentatorFollowUnit("target");
			end
		end
	elseif ( event == "AL_SPEC_PLAYER_UPDATE" ) then
		BGLive:UpdateSideFrames();
	elseif ( event == "BATTLEGROUND_POINTS_UPDATE" ) then
		BGLiveScoreBoard:UpdatePoints();
	elseif ( event == "UPDATE_BATTLEFIELD_STATUS" ) then
		local status, mapName = GetBattlefieldStatus(filter);
		print(mapName);
		local isSpectator, gameType = IsSpectator();
		local winner = GetBattlefieldWinner();
		if ( status == "active" and filter == self.currentID and winner ) then
			local database = ArenaLive:GetDBComponent(ArenaLiveSpectator.name);
			if ( ( winner == 1 and self.swapTeamNames ) or winner == 0 ) then
				database.TeamB.Score = database.TeamB.Score + 1;
			elseif ( ( winner == 0 and self.swapTeamNames ) or winner == 1 ) then
				database.TeamA.Score = database.TeamA.Score + 1;
			elseif ( winner == 255 ) then
				database.TeamB.Score = database.TeamB.Score + 1;
				database.TeamA.Score = database.TeamA.Score + 1;
			end
			BGLiveScoreBoard:UpdateTeamScore("TeamA");
			BGLiveScoreBoard:UpdateTeamScore("TeamB");
		elseif ( status == "active" and isSpectator and gameType == "battleground" ) then
			self:Enable(mapName, filter);
		elseif (  status == "none" and filter == self.currentID ) then
			self:Disable();
		end
	elseif ( event == "WORLD_STATE_UI_TIMER_UPDATE" ) then
		if ( not self.worldStateTimerIndex ) then
			for index = 1, GetNumWorldStateUI() do
				local uiType = GetWorldStateUIInfo(index);
				if ( uiType == 3 ) then -- 3 seems to be the identifier for timers.
					self.worldStateTimerIndex = index;
					break;
				end
			end
		end
		BGLiveScoreBoard:UpdateTimer();
	elseif ( event == "UPDATE_WORLD_STATES" ) then
		BGLiveScoreBoard:UpdatePoints();
	end
end

ArenaLive:ConstructAddon(BGLive, addonName, true, BGLive.defaults, false, "BGLIVE_DATABASE");
BGLive:RegisterEvent("ADDON_LOADED");
BGLive:RegisterEvent("UPDATE_BATTLEFIELD_STATUS");
